"""
Uses a modified RBF partion of unity (PU) differentiation to approximate 
L1 = u_x + u_y and L2 = u_{xx} + u_{yy} on a cicular domain

The NSPD method is used to select the shape parameter on each patch
and MDI regularization is used in forming the local differentiation matrices
for each patch

usage: run differentiationSarraPU.py

"""


from numpy.linalg import norm
from numpy import inf, zeros
import rbfCenters as rc
from rbfClassX import rbfx
from iqx import *
import warnings                     # ignore warnings of ill-conditioning
warnings.filterwarnings('ignore')

SHAPESPD = True           # either constant shape or NSPD selection
N = 1500                  # number of centers
xc, yc = rc.circleCenters(N,ch=3,center=(0.5,0.5))
N, X = len(xc), list( zip(xc,yc) )

from F2b import *   # derivative of this function being approximated
fn = F2b()
f = fn.F(xc,yc)
fDiv, fLap = fn.div(xc,yc), fn.L(xc,yc)  # exact derivatives

phi = iqx()            # inverse quadratic RBF

import rbfPU as PU
from W32 import *
phiC = W32()           # Wendland compactly supported RBF

s = 1.8                # same shape of each patch if SHAPESPD = False
radius, x, y, st = PU.cover2d(xc,yc,5,0.75,0)
n = len(st)                                     # number of patches
w = PU.patchWeights(st,x,y,xc,yc,radius,phiC)

# different shape of each patch via SPD shape parameter selection
if SHAPESPD: sp = PU.patchShape(X,st,phi,mu=0,s0=0.5,ds=0.01)

div, lap = [], []
for k in range(n):
	st0 = st[k]
	r, rx, ry = rbfx.distanceMatrix2d(xc[st0],yc[st0])
	if SHAPESPD:
		B, H, H2 = phi.rbf(r,sp[k]), phi.div(r,sp[k],rx,ry), phi.L(r,sp[k])
	else:
		B, H, H2 = phi.rbf(r,s), phi.div(r,s,rx,ry), phi.L(r,s)
	div.append( phi.dm( B, H, 5e-15, False ) )  # local differentiation matrices
	lap.append( phi.dm( B, H2, 5e-15, False ) )

# loop base approach to derivative approximation	
fad, fal = zeros((N,)), zeros((N,))
for k in range(n):
	st0, w0 = st[k], w[k]
	fad[st0] += w0*( div[k]@f[st0] )
	fal[st0] += w0*( lap[k]@f[st0] )
	
print('Sarra PU divergence error: {:1.15e}'.format( norm( fad - fDiv, inf )))
print('Sarra PU Laplacian error: {:1.15e}'.format( norm( fal - fLap, inf )))

# matrix-vector approach to derivative approximation

Ddiv, Dlap = PU.globalDMSarra(N,st,w,div), PU.globalDMSarra(N,st,w,lap)
fadAlt, falAlt = Ddiv@f, Dlap@f

print(' ')	
print('Sarra PU divergence error, DM: {:1.15e}'.format( norm( fadAlt - fDiv, inf )))
print('Sarra PU Laplacian error, DM: {:1.15e}'.format( norm( falAlt - fLap, inf )))
print(' ')
sm = 0
for k in range(n):
    sm += len(st[k])
print('average number of centers per patch: {:d}'.format(round(sm/n)))
