"""
Solves a 2d diffusion equation with a forcing term

u_t = u_{xx} + u_{yy} + 4k^2(x^2 + y^2) \cos(k \left[ t - x^2 -y^2 \right]) - 5 k \sin(k \left[t - x^2 -y^2 \right])

in a circular domain with a global RBF method.  The initial conditions
and Dirichlet boundary conditions are taken from the exact solution

u(x,y,t) = 1 + \cos(k \left[t - x^2 - y^2 \right])

usage: run heatCircleGlobal.py

see also: heatCircleGlobalEws.py which calculates the eigenvalues of the
          discretized equation with zero Dirichlet BCs applied.
          
"""

import matplotlib.pyplot as plt
from numpy.linalg import norm
from numpy import zeros_like, sin, cos, arange, inf
import matplotlib.tri as tri
import AnimatePlot as ap
import rbfCentersLib as cl
import rbfMisc as rm
import rbfCenters as rc

import warnings
warnings.filterwarnings('ignore')

from rbfClassX import rbfx
from gax import *
phi = gax()

N, Nb = 6000, 350
bpi = arange(Nb)
xi, yi = rc.circleCenters(N,cluster=False,ch=3,R=1,plt=False)
x, y = rc.circleBndCenters(xi,yi,1,Nb,tol=0.0125,plt=True)
x += 1
y += 1

print('boundary centers: {:d}'.format(Nb))
print('   total centers: {:d}'.format(len(x)))

triang = tri.Triangulation(x, y)
t, dt, finalT = 0, 0.00004, 0.5
showPlot = 1000

shape = 13.5
r, rx, ry = phi.distanceMatrix2d(x, y)
B = phi.rbf(r,shape)

mu, safe = 5e-15, False
H = phi.L(r,shape)
D2 = phi.dm(B, H, mu, safe)

K = 2
def exactSolution(x,y,t): return 1.0 + cos(K*t - K*x**2 - K*y**2)

U = exactSolution(x,y,0)
u = zeros_like(U)

def F(v,t):
    v[bpi] = exactSolution(x[bpi],y[bpi],t)
    return D2@v + 4*K**2*(x**2 + y**2)*cos(K*t - K*x**2 - K*y**2) - 5*K*sin(K*t - K*x**2 - K*y**2)
    
def makePlot(): 
	plt.tricontour(triang, u, 15, linewidths=1.5, colors='k')
	plt.tricontourf(triang, u, 15, cmap=plt.cm.jet)
	

n = 0
while t<finalT:
    u = rm.rk4(U,t,dt,F)
    u[bpi] = exactSolution(x[bpi],y[bpi],t)
    if n % showPlot == 0:
        ap.animate(makePlot)
    t += dt
    n += 1
    U = u.copy()

plt.colorbar()
exact = exactSolution(x,y,t)
print('max error: {:1.2e}'.format(norm( u-exact,inf)/max(abs(u))))

