""" Module for working with RBF matrices that have a centrosymmetric or
skew-centrosymmetric struncture.

Reference:  S. A. Sarra. "Radial Basis Function Methods - the case of symmetric domains".
            Numerical Methods for Partial Differential Equations, 2018. DOI:10.1002/num.22272
			
import suggestion: import rbfCentro as cn

--------------------------------- function summary -------------------------------

tests for symmetry:
	* isCentro
	* isSkewCentro
	* hasSymmetry

fullCentroMatrix - "corrects" a matrix that theoretically should have symmetry but was
                   formed by a "non-symmetric" algorithm
centroDecomposeMatrix - returns the smaller L and M matrices
centroMult - matrix/vector multiplication   
centroEigenvalues - eigenvalues
centroConditionNumber - 2 norm condition number
solveCentro - solution of linear systems
centroDM - constucts differentiation matrices
             	
production of centrosymmetric center distributions in 2d domains:
	* centroCircle
	* centroCenters

"""

from numpy import amax, fliplr, flipud, zeros, shape, arange, append, eye, diag_indices_from, nonzero, ravel
from rbfClassX import rbfx
from scipy.linalg import svdvals, solve, lu_factor, lu_solve, cho_factor, cho_solve, eigvals
import matplotlib.pyplot as pp
import rbfCenters as rc


# -------------- centrosymmetry tests -----------------------------------------

"""

Note: In testing a matrix for centro symmetry, the full N x N matrix constructed
      Non-centrosymmetric algorithms must be used as extending an arbitrary 
      N x (N/2) matrix to an N x N matrix with rbfCentro.fullCentroMatrix will make
      the full matrix centrosymmetric

isCentro and isSkewCentro - test a matrix for  (skew)centrosymmetry

  input
     B    a N x N matrix

  output
     s    returns possibly zero or more probably a very small number, e.g. 1e-15, 
          for a N x N centrosymmetric input. otherwise, the return of a larger
          number indicates the input matrix isnot centrosymmetric
"""


def isCentro(B): return amax( abs(B - fliplr(flipud(B))) )

def isSkewCentro(B): return amax( abs(B + fliplr(flipud(B))) )

def hasSymmetry(B,small=1e-15):
	""" tests a N x N matrix for both centrosymmetry and skew-centrosymmetry
	
	input
	   B    a N x N matrix
	
	output
	   prints the result and returns:
		  1   skew
		  2   centro
		  0   neither
		
	
	example usage:
	"""
	
	c1, c2 = isCentro(B), isSkewCentro(B)
	if c1 < small: 
		print('centrosymmetric')
		return 2
	elif c2 < small:
		print('skew centrosymmetric')
		return 1
	else:
		print('no centrosymmetry: c1 = {0:1.2e}, c2 = {1:1.2e}'.format(c1,c2))
		return 0
		
# ----------------------------------------------------------------------------

def fullCentroMatrix(Dh,N,skew):
	"""  Corrects a full centro matrix constructed with a standard algorithm so that is has
	 the correct symmetry or takes the left half of a matrix and expands it to have symmetry
	
	 inputs
	  Dh      either a N x N  centro or skew-centro symmetric matrix
				or the or the (N) x (N/2) left half of a such a matrix
	 skew    1 (or any positve odd integer) if skew-centro
			 2 (or any positive even integer) if centro
	   N     N must be even
	
	outputs
	   D     the full N x N matrix
	"""
	n = int(N/2)
	D = zeros((N,N))
	D[:,:n] = Dh[:,:n]                                                #  left half
	D[:n,n:] = (-1.0)**(skew)*fliplr( flipud(  D[n:,:n] ) )     #   D_{12}
	D[n:,n:] = (-1.0)**(skew)*fliplr( flipud(  D[:n,:n] ) )     #   D_{22}
	return D
		
# ------------- matrix vector multiplication ---------------------------------

def centroDecomposeMatrix(D,rho):
	"""  A (skew) centrosymmetric matrix is similar to a 
							block diagonal matric
							 [ L   O ]
							 [ 0   M ]
	 Given D, this function computes L and M which are used in condition
	 number, eigenvalue, and multiplication algorithms.
	
	 inputs
	 D      either a N x N  centrosymmetric matrix DM
			or the or the (N) x (N/2) left half of a centrosymmetric matrix DM
	 rho    1 if D is skewcentrosymmetric
			2 if D is centrosymmetric
	
	 outputs
	   L/2    the even (N/2) x (N/2) DM
	   M/2    the odd  (N/2) x (N/2) DM
	
	 example usage: 
	"""
	N = shape(D)[0]
	N2 = int(N/2)
	a = D[:N2,:N2]                                # B_{11}
	b = (-1.0)**(rho+1)*flipud( D[N2:,:N2] )   # J x B_{21}
	L = 0.5*(a - b)   # L = 0.5*(B_{11} - J*B_{21}) or Lh = 0.5*( B_{11} + J*B_{21})
	M = 0.5*(a + b)   # M = 0.5*(B_{11} + J*B_{21}) or Mh = 0.5*( B_{11} - J*B_{21} )
	return L, M
	

def centroMult(u,L,M,rho):
	""" Matrix-vector multiplication with a (skew)centrosymmetric matrix
	 
	inputs
	  u    N x 1 vector
	  L    the even (N/2) x (N/2) DM  (L and M from rbfCentro.centroDecomposeMatrix)
	  M    the odd  (N/2) x (N/2) DM
	 rho    1 if B (the marix from which L and M are derived from) is skewcentrosymmetric 
			2 if B is centrosymmetric
	
	outputs
	 ua    N x 1 vector that is the results of D*f
	
	example usage:
	"""
	N = len(u)
	N2 = int(N/2)
	k = arange(N2)
	
	t = flipud( u[N2:] )	               # decompose
	e, o = u[:N2] + t, u[:N2] - t          #  xe_1 and xo_1
	
	uae, uao = L@e, M@o                    # fe_1, fo_1
	
	ua = zeros((N,))                       # reconstruct
	ua[:N2] = uae + uao                        #     f_1 = fe_1 + fo_1
	s1, s2 = (-1.0)**(rho), (-1.0)**(rho+1)	
	ua[N2:] = flipud( s1*uae + s2*uao )    # f_2 = s1*J*fe_1 + s2*J*fo_1
	
	return ua

# ------------------------------------------------------------------------

def centroEigenvalues(B):
	L, M = centroDecomposeMatrix(B,0)
	ewL, ewM = eigvals(2*L), eigvals(2*M)
	return append(ewL, ewM)
	
	
# ----------- condition numbers -------------------------------------------

def centroConditionNumber(B,mu):
	""" inputs
	   B    N x (N/2) left half (or full N x N) centro system matrix.
			Only the left half is used.
	  mu    MDI regularization parameter
	
	outputs:
	   kappaL
	   kappaM
	   kappaB   2-norm condition number of the matrix B
	
	example usage:
	"""
	L, M = centroDecomposeMatrix(B,0)
	N = shape(L)[0]
	I = eye(N)
	sL, sM = svdvals(2*(L + mu*I)), svdvals(2*(M + mu*I))
	kappaL, kappaM = sL[0]/sL[-1], sM[0]/sM[-1]
	s = append(sL, sM)
	kappaB = max(s)/min(s)
	return kappaB, kappaL, kappaM


# ----------- solution of linear systems ----------------------------------

def solveCentro(B,f,mu=5e-15,safe=True):
	""" Solves a centrosymmetric linear system Ba = f.
	
	 inputs
		B      either a N x N  centrosymmetric matrix (N must be even)
			   or the or the (N) x (N/2) left half of a centrosymmetric matrix
		f      N x 1 vector
		mu     MDI regularization parameter (mu=0 for no regularization)
		safe  true -> use backslash operator
			   false -> directly use Cholesky factorization
	
	outputs
		a    N x 1 solution of B a = f
	
	example usage:
	"""
	
	N = len(f)
	n = int(N/2)
	A = B[:n,:n]                         # B_{11}
	t1 = flipud( B[n:,:n] )           # J x B_{21)  
	L, M = A - t1, A + t1                 # B_{11} - J x B_{21}, B_{11} + J x B_{21}
	
	if mu>0:
		L[diag_indices_from(L)] += mu    # MDI regularization
		M[diag_indices_from(M)] += mu
		
	b1 = f[:n]
	t2 = flipud( f[n:] )                 # t2 = J x b2
	b1p, b2p = b1 - t2, b1 + t2
	
	if safe:
		(luL, pivL), (luM, pivM) = lu_factor(L), lu_factor(M)
		x1h, x2h = lu_solve((luL, pivL), b1p), lu_solve((luM, pivM), b2p)
	else:
		(cL, lowL), (cM, lowM) = cho_factor(L), cho_factor(M)
		x1h, x2h = cho_solve((cL, lowL), b1p), cho_solve((cM, lowM), b2p)
	
		
	return append( 0.5*( x1h + x2h ), 0.5*flipud(x2h - x1h) ) 
	
	
# ----------- derivative matrix formation ---------------------------------

def centroDM(B,F,N,rho,mu=5e-15,safe=True):
	""" Constructs a RBF differentiation matrix.
	
	inputs
	   B    N x (N/2) left half (or full N x N) centro system matrix.
			Only the left half is used.
	   F    N x (N/2) left half (or full N x N) derivative evaluation matrix.
			Only the left half is used.
	   N    Number of columns in B and F.  N must be even.
	 rho    1 if B is skewcentrosymmetric
			2 if B is centrosymmetric
	  mu    MDI regularization parameter (optional).  
	 safe   True (LU with partial pivoting), False (Cholesky)
	
	outputs
	   D    N x (N/2) left half of the DM. If the full DM is needed it can be
			constructed with rbfCentro.fullCentroMatrix.
	
	example usage:
	"""
	F *= (-1)**(rho)
	n = int(N/2)
	D = zeros((N,n))
	A = B[:n,:n]           # B_{11}
	t1 = flipud( B[n:,:n] )   # J x B_{21}
	L, M = A - t1, A + t1
	
	if mu>0:
		L[diag_indices_from(L)] += mu    # MDI regularization
		M[diag_indices_from(M)] += mu
		
	b1 = F[:n,:n]
	t2 = flipud( F[n:,:n] )
	b1p, b2p = b1 - t2, b1 + t2
	
	if safe:
		(luL, pivL), (luM, pivM) = lu_factor(L), lu_factor(M)
		x1h, x2h = lu_solve((luL, pivL), b1p), lu_solve((luM, pivM), b2p)
	else:
		(cL, lowL), (cM, lowM) = cho_factor(L), cho_factor(M)
		x1h, x2h = cho_solve((cL, lowL), b1p), cho_solve((cM, lowM), b2p)

		
	D[:n,:n], D[n:,:n] = 0.5*( x1h + x2h ).T, (-1.0)**(rho)*fliplr( flipud( 0.5*flipud(x2h - x1h).T ) )      # WTF ????
	return D 
	

	
# ------- centrosymmetric center distributions in 2d domains --------------



def centroCircle(N,cluster,ch,R=1,plt=False):
	""" centroCenters for the specific cace of a circle. Quasirandom 
				   centers on a circle of radius R with centro placement. The centers
					are either based on a Hammersley or Halton sequence.
	  inputs
		 N      Number of centers in the covering square.  The number of centers
				returned is less than N.
	   cluster  Logical variable for clustering option
		ch          1        Halton
				otherswise   Hammersley
		 R      Radius of the circle
		plt     Logical variable for plotting option
	
	 outputs
	   xc, yc     center coordinates
	
	 example usage:
	"""
	x, y = rc.circleCenters(N,cluster,ch,R,False)
	I, = nonzero(np.ravel(y>x))
	xc = append( x[I], flipud(-x[I]) )
	yc = append( y[I], flipud(-y[I]) )
	if plt: pp.scatter(x,y,c='g',marker='.')
	return xc, yc
	
	
def centroCenters(x,y,symType,plt=False):     # NOT TESTED YET
	""" inputs
	   x, y      centers covering an entire domain
	 symType     0 - y axis
				 1 - x axis
				 2 - origin
	   plt       logical variable to plot centers
	
	outputs      
	  xc, yc      centrosymmetric center distribution
	
	example usage:
	"""
	
	if symType == 0:
		I, = nonzero(ravel(x>0))
		xc = append( x[I], flipud(-x[I]) )
		yc = append( y[I], flipud(y[I]) )
	elif symType == 1:
		I, = nonzero(ravel(y>0))
		xc = append( x[I], flipud(x[I]) )
		yc = append( y[I], flipud(-y[I]) )
	else:
		I, = nonzero(ravel(y>x))
		xc = append( x[I], flipud(-x[I]) )
		yc = append( y[I], flipud(-y[I]) )
		
	if plt: pp.scatter(x,y,c='g',marker='.'), pp.show()
	
	return xc, yc
	



# -------------------------------------------------------------------------

		
















