"""
A collection of miscellaneous functions for working with RBF methods.

suggested import: import rbfMisc as RM

"""

from numpy import sqrt, real, imag, meshgrid, linspace

# -----------------------------------------------------------------------

def rk4(v,t,k,F):
	""" Fourth-order Runge-Kutta ODE method. Advances an ODE system from
        time t^n to time t^{n+1} with constant step size k
	
	input
	
	  V  solution at time t^n
	  t  time t^n
	  k  time step size delta t
	  F  the F of the ODE system v_t = F[v,t]
	
	output
	  v  solution at time t^{n+1}
	"""
	F1 = F(v, t)
	F2 = F(v + 0.5*k*F1, t + 0.5*k)
	F3 = F(v + 0.5*k*F2, t + 0.5*k)
	F4 = F(v + k*F3, t + k)
	return v + k*(F1 + 2.0*F2 + 2.0*F3 + F4)/6.0


# ---------------------------------------------------------------------

def longSideMask(x, y, triangles, threshold=0.5):
    triOut = []
    for centers in triangles:
        a,b,c = centers
        side0 = sqrt( (x[a] - x[b]) **2 + (y[a] - y[b])**2 )
        side1 = sqrt( (x[b] - x[c]) **2 + (y[b] - y[c])**2 )
        side2 = sqrt( (x[c] - x[a]) **2 + (y[c] - y[a])**2 )
        maxLength = max([side0, side1, side2])
        if maxLength > threshold:
            triOut.append(True)
        else:
            triOut.append(False)
    return triOut


# ----------------------------------------------------------------------

from scipy.sparse import issparse
from scipy.linalg import eigvals
import matplotlib.pyplot as plt

def dmEWs(D,dt=1,p1=True,p2=True,verbose1=True,verbose2=False):
	""" Finds all ews of a dense or sparse differentiation matrix.
	    Option to plot RK4 stability region and ews scaled by dt.
	"""
	if issparse(D):
		ews = eigvals(D.todense())
	else:
		ews = eigvals(D)
		
	ews *= dt
	
	if verbose1:
		print('max real ew: {:1.3e}'.format(max(real(ews))))
	if verbose2:
		print('max imag ew: {:1.3e}'.format(max(imag(ews))))
		print('min real ew: {:1.3e}'.format(min(real(ews))))
	
	if p1:
		fig = plt.figure()
		ax = fig.add_subplot(111)
		ax.plot( real(ews), imag(ews), 'b.')
	if p2:
		x, y = meshgrid(linspace(-3.5, 0.5, 200), linspace(-3.5, 3.5, 200))
		z = x + y*1j
		R = absolute(1 + z + 0.5*z**2 + (z**3)/6.0 + (z**4)/24.0)
		ax.contour(x,y,R,[1,1.001],colors='g')
		plt.xlabel('Re(z)')
		plt.ylabel('Im(y)')
		ax.set_aspect('equal')
	
	plt.show()
	
	
# -----------------------------------------------------------------------



from scipy.sparse.linalg import eigs
from numpy import absolute, concatenate

def dmSparseEWs(D,k=2,dt=1,p1=True,p2=True):
	""" Finds the k ews with the following properties of a sparse DM:
	    largest positive real part, smallest real part, and
		the largest imaginary part
		
		p1 = True -> plot ews
        p2 = True -> plot RK4 stability region
		
	"""
	ews1 = eigs(D,k,which='LR',return_eigenvectors=False)
	ews2 = eigs(D,k,which='SR',return_eigenvectors=False)
	ews3 = eigs(D,k,which='LI',return_eigenvectors=False)
	ews = concatenate([ews1,ews2,ews3])
	ews *= dt
	print('max imag ew: {:1.3e}'.format(max(sp.imag(ews))))
	print('min real ew: {:1.3e}'.format(min(sp.real(ews))))
	print('max real ew: {:1.3e}'.format(max(sp.real(ews))))
	if p1:
		fig = plt.figure()
		ax = fig.add_subplot(111)
		ax.plot( real(ews), imag(ews), 'b.')
	if p2:
		x, y = meshgrid(linspace(-3.5, 0.5, 200), linspace(-3.5, 3.5, 200))
		z = x + y*1j
		R = absolute(1 + z + 0.5*z**2 + (z**3)/6.0 + (z**4)/24.0)
		ax.contour(x,y,R,[1,1.001],colors='g')
		plt.xlabel('Re(z)')
		plt.ylabel('Im(y)')
		ax.set_aspect('equal')
	
	plt.show()
	








