"""
Module for implementing RBF partion of unity (PU) methods.

Two versions of the method are implemented:
   1) the "standard" PU method
   2) a modified PU (MPU) method
    
   The MPU was first used in reference:
   
   S. A. Sarra and Y. Bai.  "A rational radial basis function method for accurately resolving
   discontinuities and steep gradients".  Applied Numerical Mathematics 130, p. 131-142, 2018.
   DOI: 10.1016/j.apnum.2018.04.001 

import suggestion: import rbfPU as PU


functions:
	1) centersInCircles - determines which centers are in a patch
	2) cover2d - covers a 2d region with overlapping circles
	3) patchShape - uses NSPD criteria to select a shape parameter on each patch
	4) patchWeights - 
	5) patchWeightsD
	6) evaluateDerivatives
	7) globalDM
	
	6b) evaluateDerivativesSarra
	7b) globalDMSarra
	

graphical functions:
	1) plotCover
	2) plotOneStencil



"""

from numpy import shape, sqrt, where, linspace, delete, zeros, array, in1d, ones, nonzero, meshgrid
from numpy import sin, cos, pi, ravel
import matplotlib.pyplot as pp
from scipy.sparse import *
from rbfClassX import rbfx


# -------------------------------------------------------------------------------------------

def centersInCircles(xc,yc,x0,y0,R):
	""" Given RBF centers in a 2d domain determine which centers lie
	    in a circle with center (x0,y0) and radius R.
		
	Inputs:
	  xc, yc   RBF centers
	  x0, y0   center of the circle
	    R      radius of the circle
		
	Output:
	   st      indicies of the RBF centers that lie within the circle
	
	"""
	nc = shape(x0)[0]
	st = []
	xc, yc = xc.flatten(), yc.flatten()
	for k in range(nc):
		r = sqrt( ( xc-x0[k] )**2 + ( yc -y0[k] )**2 )
		I = where( r<=R )
		st.append(I[0])
	return st
# -------------------------------------------------------------------------------------------



def cover2d(xc,yc,m,fac=0.75,threshold=8,verbose=True):
	""" Covers a 2d region with a m x m distribution of overlapping circles.
	
	Inputs:
	  xc, yc    RBF centers
	    m       number of circles in the x and y directions
	   fac      controls the amount of circle overlap, recommended range 0.75 <= fac <= 1.0
	               fac < 0.75 may lead to some centers not being in a cover
				   fac = 0.5 -> neighboring circles only touch and to not overlap at all
				   fac = 1.0 -> given 3 consecutive circles, the middle circle will overlap half its neighbor
  threshold     delete covering circles with less than threshold centers
    verbose     True -> print the average number of centers per patch
  
    Outputs:
	  x, y      the centers of the covering circles
	 radius     the radius of the circles (all the same)
	
	"""
	xMin, xMax = min(xc), max(xc)
	yMin, yMax = min(yc), max(yc)
	
	my = m                # number of covering circles in the y direction
	delta = (yMax - yMin)/my
	mx = int(round((xMax - xMin)/delta))
	x = linspace(xMin + 0.5*delta,xMax - 0.5*delta,mx)
	y = linspace(yMin + 0.5*delta,yMax - 0.5*delta,my)
	[x, y] = meshgrid(x,y)                            # circle centers
	x, y = x.flatten(), y.flatten()
	radius = fac*delta
	st = centersInCircles(xc,yc,x,y,radius)

	M = len(st)
	deleteIndex = []
	for k in range(M-1,-1,-1):
		if len(st[k])<threshold:
			deleteIndex.append(k)
			del st[k]
			
	x, y = delete(x,deleteIndex), delete(y,deleteIndex)
	
	n = len(st)
	sm = 0
	for k in range(n):
		sm += len(st[k])
	avg = round(sm/n)
	print('average number of centers per patch: {:d}'.format(avg))
	
	return radius, x, y, st


# -----------------------------------------------------------------

def patchShape(X,st,phi,mu=0,s0=0.5,ds=0.01):
	""" Based on the system matrix being numerically positive definite,
	    determines the shape parameter to be used on each patch of a cover.
		
	Inputs:
	   X   RBF centers in the format  X = list( zip(x,y) )
	   st  the indices of the RBF centers on each patch
	  phi  SPD RBF
	   mu  method of diagonal increments regularization parameter
	   s0  initial shape parameter for the first patch
	   ds  increment in which to change the shape parameter
	   
	Output:
	    s  an array of shape parameters for the cover.
	"""
	P = len(st)              # number of patches
	initS = s0
	s = zeros((P,))
	x, y = zip(*X)
	x, y = array(x), array(y)
	for k in range(P):
		x0 = list( zip(x[st[k]],y[st[k]]) )
		s[k], _ = rbfx.spdShape(phi,x0,initS,ds,mu)
		initS = s[k]        # initial shape on next patch = final shape on previous patch
		
	return s

# -------------------------------------------------------------------------------------------


#  p     psi_k(x)
#  ps    sum_{i=1}^P psi_i(x)

def patchWeights(st,x0,y0,x,y,radius,phi):   
	""" Patch weights for standard PU interpolation and both interpolation and
	    derivative for the Sarra PU (SPU) approach.
		
	Inputs:
	  st    indices of RBF centers in each covering circle
	x0, y0  the centers of each covering circle
	 x, y   the RBF centers
   radius   the radius of the covering circles
      phi   a compactly supported RBF
	  
	Output:
	  w     the patch weights
	
	"""
	N = len(x0)              # number of covering patches
	w = []                   # weights on each patch
	for i in range(N):       # loop over each patch
		st0 = st[i]          # indicies of current patch
		n = len(st0)
		w0 = ones(n)         # each weight is initially 1 => center located in only one patch
		
		for k in range(n):    # loop over elements of the current patch and ckeck if they are
			ps = 0         #      located in any other patches, if so adjust the weight
			for j in range(N):   # check if st0[k] is in patch j
				if j!=i:         # this case already taken care of by setting w = 1 initially
					if in1d(st0[k],st[j]):
						xs, ys = x[st0[k]] - x0[j], y[st0[k]] - y0[j]
						ps += phi.rbf( sqrt(xs**2 + ys**2), radius)		
				if ps>0:          # the point is in more than one patch
					xs, ys = x[st0[k]] - x0[i], y[st0[k]] - y0[i]
					psiK = phi.rbf( sqrt(xs**2 + ys**2), radius)
					w0[k] = psiK/(ps + psiK)
		w.append(w0)	
	return w

# ----------------------------------------------------------------------------------------
# -------------- Sarra's PU method -------------------------------------------------------
# ----------------------------------------------------------------------------------------


def globalDMSarra(N,st,w,Ds):
	""" Constructs the global differentiation matrix for the Sarra PU (SPU) method
		for any order linear differential operator.
		
	Inputs:
	   N   the total number of RBF centers
	  st   indices of the RBF centers in each covering circle
	   w   patch weights
	  Ds   a list containing local differentiation matrices for each patch
	  
	Output:
	  D    a sparse, global, differentiation matrix
	"""
	D = lil_matrix((N,N))    # efficient format for sparse matrix construction
	P = len(st)              # number of patches
	
	for j in range(N):       # loop over each row
		for i in range(P):   # loop over each patch
			st0 = st[i]             # indicies of current patch
			w0 = w[i]               # weights for the current patch
			d0 = Ds[i]              # local DM for the patch
			n = len(st0)
			if in1d(j,st0):      # if x[j] is in the patch, update the cols in the row
				I = where(st0==j)
				for k in range(n):
					D[j,st0[k]] += w0[I]*d0[I,k] 
	return D.tocsr()

# -------------------------------------------------------------------------------------------

def evaluateDerivatesSarra(f,st,D,w):
	""" Evaluates derivative for the Sarra PU (SPU) method without forming a
	    global differentiation matrix
		
	Inputs:
	   f   function values at the RBF center locations of the function being differentiated
	  st   indices of the RBF centers in each covering circle
	   w   patch weights
	   D   a list containing local differentiation matrices for each patch
	  
	Output:
	  fa  the approximate derivative values at the RBF center locations
	   
	"""
	N, P = len(f), len(st)
	fa = zeros(N)
	for k in range(P):
		st0, w0 = st[k], w[k]
		fa[st0] +=  w0*( D[k]@f[st0] )
	return fa
		
				
# ------------------------------------------------------------------------------------------
# --------- "standard" PU differentiation --------------------------------------------------
# ------------------------------------------------------------------------------------------

#def globalDM(N,st,w,wp,Ds1,wpp,Ds2):
def globalDM(N,st,w,wp,Ds1,Ds2): 
	""" Constructs the global differentiation matrix for the standard PU (PU) method
		for first and second order derivatives.
		
	Inputs:
	   N   the total number of RBF centers
	  st   indices of the RBF centers in each covering circle
	   w   patch weights
	  wp   first derivative of the patch weights
	 Ds1   a list containing local first-order differentiation matrices for each patch
	 Ds2   a list containing local second-order differentiation matrices for each patch
	  
	Output:
	  D, D2    sparse, global, first and second order differentiation matrices
	"""
	D, D2 = lil_matrix((N,N)), lil_matrix((N,N))
	P = len(st)              # number of patches
	
	for j in range(N):       # loop over each row
		for i in range(P):   # loop over each patch
			st0 = st[i]             # indicies of current patch
			n = len(st0)            # size of the current patch
			w0, wp0 = w[i], wp[i]   # weights for the current patch
			d01, d02 = Ds1[i], Ds2[i]             # local DM for the patch
			if in1d(j,st0):      # if x[j] is in the patch, update the cols in the row corresponding to x[j]
				I, = nonzero(ravel(st0==j))
				for k in range(n):
					D[j,st0[k]] += w0[I]*d01[I,k]
					D2[j,st0[k]] += w0[I]*d02[I,k] + 2*wp0[I]*d01[I,k]
			
	return D.tocsr(), D2.tocsr()

# -------------------------------------------------------------------------------------------

def evaluateDerivates(f,st,D1x,D1y,D2x,D2y,w,wp,wpp,wpy,wppy,INCLUDEZEROTERMS=False):
	""" Evaluates first and second order derivatives for the standard PU (PU) method
	    without forming a global differentiation matrix
		
	Inputs:
	   f   function values at the RBF center locations of the function being differentiated
	  st   indices of the RBF centers in each covering circle

	   D   a list containing local differentiation matrices for each patch
	     
		  PATCH WEIGHTS AND DERIVATIVES
	   w   weights
	  wp   first derivative wrt x
	 wpp   second derivative wrt x
	 wpy   first derivative wrt y
	wppy   second derivative wrt y
	 
	     LOCAL DIFFERENTATION MATRICES FOR EACH PATCH
	 D1x   first-order wrt x
	 D2x   second-order wrt x
	 D1y   first-order wrt y
	 D2y   second-order wrt y
	  
	Output: approximate derivative values at the RBF center locations
	  fx    first derivative wrt x
	  fy    first derivative wrt y
	 fxx    second derivative wrt x
	 fyy    second derivative wrt x
	 
	 INCLUDEZEROTERMS  Include terms that are theorectically zero.  In floating point
	                   arithemetic the terms will ussually not be zero due to rounding errors
					   but they will be relatively small.
	   
	"""
	N, P = len(f), len(st)
	fx, fy, fxx, fyy = zeros(N), zeros(N), zeros(N), zeros(N)
	
	if INCLUDEZEROTERMS:
		for k in range(P):
			st0, w0, wp0, wpp0, wpy0, wppy0 = st[k], w[k], wp[k], wpp[k], wpy[k], wppy[k]
			fx[st0] +=  w0*( D1x[k]@f[st0] ) + wp0*f[st0]  
			fy[st0] +=  w0*( D1y[k]@f[st0] ) + wpy0*f[st0]  
			fxx[st0] += w0*( D2x[k]@f[st0] ) + 2*( (D1x[k]@f[st0])*wp0  ) + f[st0]*wpp0 
			fyy[st0] += w0*( D2y[k]@f[st0] ) + 2*( (D1y[k]@f[st0])*wpy0 ) + f[st0]*wppy0
	else:
		for k in range(P):
			st0, w0, wp0, wpp0, wpy0, wppy0 = st[k], w[k], wp[k], wpp[k], wpy[k], wppy[k]
			fx[st0] +=  w0*( D1x[k]@f[st0] ) 
			fy[st0] +=  w0*( D1y[k]@f[st0] )   
			fxx[st0] += w0*( D2x[k]@f[st0] ) + 2*( (D1x[k]@f[st0])*wp0  ) 
			fyy[st0] += w0*( D2y[k]@f[st0] ) + 2*( (D1y[k]@f[st0])*wpy0 ) 
			
	return fx, fy, fxx, fyy



# ----------------------------------------------------------------------------------

#  p/psiK     psi_k(x)
#  ps    sum_{i=1}^P psi_i(x)

def patchWeightsD(st,x0,y0,x,y,radius,phi):   # standard PU: interp, 1st, and 2nd derivative
	
	N = len(x0)              # number of covering patches
	w, wp, wpp, wpy, wppy = [], [], [], [], []  # weights on each patch
	for i in range(N):       # loop over each patch
		st0 = st[i]          # indicies of current patch
		n = len(st0)
		w0, wp0, wpp0, wpy0, wppy0 = ones(n), zeros(n), zeros(n), zeros(n), zeros(n)           

		for k in range(n):
			ps, psx, psxx, psy, psyy = 0, 0, 0, 0, 0
			for j in range(N):   # check if st0[k] is in patch j
				if in1d(st0[k],st[j]):
					xs, ys = x[st0[k]] - x0[j], y[st0[k]] - y0[j]
					ps += phi.rbf( sqrt(xs**2 + ys**2), radius)	
					psx += phi.D1( sqrt(xs**2 + ys**2),radius,xs)
					psxx += phi.D2( sqrt(xs**2 + ys**2),radius,xs)
					psy += phi.D1( sqrt(xs**2 + ys**2),radius,ys)
					psyy += phi.D2( sqrt(xs**2 + ys**2),radius,ys)
						
			xs, ys = x[st0[k]] - x0[i], y[st0[k]] - y0[i]    # i is the current patch 
			p = phi.rbf( sqrt(xs**2 + ys**2), radius)     # psiK
			px = phi.D1( sqrt(xs**2 + ys**2),radius,xs)
			pxx = phi.D2( sqrt(xs**2 + ys**2),radius,xs)
			py = phi.D1( sqrt(xs**2 + ys**2),radius,ys)
			pyy = phi.D2( sqrt(xs**2 + ys**2),radius,ys)
			w0[k] = p/ps
			wp0[k] = ( px*ps - p*psx )/ps**2
			wpp0[k] = ( ps**2*( ps*pxx - p*psxx ) - ( 2*ps**2*px*psx -2*p*psx**2*ps) )/ps**4
			wpy0[k] = ( py*ps - p*psy )/ps**2
			wppy0[k] = ( ps**2*( ps*pyy - p*psyy ) - ( 2*ps**2*py*psy -2*p*psy**2*ps) )/ps**4
				
		w.append(w0)
		wp.append(wp0)
		wpp.append(wpp0)
		wpy.append(wpy0)
		wppy.append(wppy0)
	return w, wp, wpp, wpy, wppy



# -------------------------------------------------------------------------------------------
# ---------------------- graphics -----------------------------------------------------------
# -------------------------------------------------------------------------------------------

def plotOneStencil(x,y,coverNumber,st,x0,y0):
	""" Plots one circle of a cover.
	
	Inputs:
	   x, y    RBF center
 coverNumber   cover to plot
        st     the RBF centers in the circle being plotted
	 x0, y0    the center of the circle being plotted	
	   
	"""
	pp.scatter(x,y,c='b',marker='.')
	I = st[coverNumber]
	pp.scatter(x[I],y[I],c='g',marker='.')
	pp.scatter(x0[coverNumber],y0[coverNumber],c='r',marker='*')
	pp.show()
	
# -------------------------------------------------------------------------------------------

def plotCover(xc,yc,x,y,radius):
	""" Plots the covering circles over a 2d domain.
	
	Inputs:
	   xc, yc   RBF centers
	    x, y    the center of each covering circle
	  radius    the radius of the covering circles (all the same)	
	"""
	pp.scatter(xc,yc,c='b',marker='.')
	M = len(x)
	th = linspace(0,2*pi)
	for k in range(M):
		xa, ya = radius*cos(th) + x[k], radius*sin(th) + y[k]
		pp.plot(xa,ya,'k')
	pp.show()
	
# -------------------------------------------------------------------------------------------



